@@ -1,10 +1,12 @@
 package main
 
 import (
-	"bytes"
+	"crypto/sha1"
+	"encoding/hex"
 	"encoding/json"
 	"errors"
 	"fmt"
+	"hash"
 	"io"
 	"io/ioutil"
 	"log"
@@ -21,6 +23,12 @@ import (
 	"github.com/olekukonko/tablewriter"
 )
 
+func openTruncate(out string) (*os.File, error) {
+	base := filepath.Dir(out)
+	_ = os.MkdirAll(base, os.ModePerm)
+	return os.OpenFile(out, os.O_TRUNC|os.O_WRONLY|os.O_CREATE, os.ModePerm)
+}
+
 func resolveContentType(fileExtensionOrContentType string) runner.MimeType {
 	if strings.Contains(fileExtensionOrContentType, string(filepath.Separator)) {
 		return runner.MimeType(fileExtensionOrContentType)
@@ -29,23 +37,20 @@ func resolveContentType(fileExtensionOrContentType string) runner.MimeType {
 	return runner.GetMimeType("x."+fileExtensionOrContentType, runner.ContentTypeInfo{})
 }
 
-func openTruncate(out string) (*os.File, error) {
-	base := filepath.Dir(out)
-	_ = os.Mkdir(base, os.ModePerm)
-	return os.OpenFile(out, os.O_TRUNC|os.O_WRONLY|os.O_CREATE, os.ModePerm)
-}
-
-func evalFileInto(file string, out *os.File) error {
-	if file == "" {
-		return fmt.Errorf(`First argument when not used in a pipe should be a file. e.g. 'dsq test.csv "SELECT COUNT(1) FROM {}"'`)
-	}
-
-	mimetype := runner.GetMimeType(file, runner.ContentTypeInfo{})
+func evalFileInto(file, mimetype string, out *os.File) error {
 	if mimetype == "" {
-		return fmt.Errorf("Unknown mimetype for file: %s.", file)
+		mimetype = string(runner.GetMimeType(file, runner.ContentTypeInfo{}))
+	} else {
+		mimetype = string(resolveContentType(mimetype))
 	}
 
-	return runner.TransformFile(file, runner.ContentTypeInfo{}, out)
+	if mimetype == "" {
+		return fmt.Errorf("Unknown mimetype for file: %s.\n", file)
+	}
+
+	return runner.TransformFile(file, runner.ContentTypeInfo{
+		Type: mimetype,
+	}, out)
 }
 
 func getShape(resultFile, panelId string) (*runner.Shape, error) {
@@ -113,6 +118,7 @@ func dumpJSONFile(file string, pretty bool, schema bool) error {
 		if err != nil {
 			return err
 		}
+		// This is intentional to add a newline
 		fmt.Println()
 		return nil
 	}
@@ -162,6 +168,138 @@ func dumpJSONFile(file string, pretty bool, schema bool) error {
 	return nil
 }
 
+func getFileContentHash(sha1 hash.Hash, fileName string) error {
+	file, err := os.Open(fileName)
+	if err != nil {
+		return err
+	}
+	defer file.Close()
+
+	_, err = io.Copy(sha1, file)
+	return err
+}
+
+func getFilesContentHash(files []string) (string, error) {
+	sha1 := sha1.New()
+
+	for _, file := range files {
+		err := getFileContentHash(sha1, file)
+		if err != nil {
+			return "", err
+		}
+	}
+
+	return hex.EncodeToString(sha1.Sum(nil)), nil
+}
+
+func importFile(projectId string, file, mimetype string, ec runner.EvalContext) (*runner.PanelInfo, error) {
+	panelId := uuid.New().String()
+	resultFile := ec.GetPanelResultsFile(projectId, panelId)
+	out, err := openTruncate(resultFile)
+	if err != nil {
+		return nil, err
+	}
+	defer out.Close()
+
+	if err := evalFileInto(file, mimetype, out); err != nil {
+		return nil, err
+	}
+
+	s, err := getShape(resultFile, panelId)
+	if err != nil {
+		return nil, err
+	}
+
+	return &runner.PanelInfo{
+		ResultMeta: runner.PanelResult{
+			Shape: *s,
+		},
+		Id:   panelId,
+		Name: uuid.New().String(),
+	}, nil
+}
+
+type args struct {
+	pipedMimetype string
+	pretty        bool
+	schema        bool
+	sqlFile       string
+	cacheSettings runner.CacheSettings
+	nonFlagArgs   []string
+	dumpCacheFile bool
+}
+
+func getArgs() (*args, error) {
+	args := &args{}
+	osArgs := os.Args[1:]
+	for i := 0; i < len(osArgs); i++ {
+		arg := osArgs[i]
+		isLast := i == len(osArgs)-1
+
+		if arg == "--verbose" {
+			runner.Verbose = true
+			continue
+		}
+
+		if arg == "-s" || arg == "--stdin" {
+			if isLast {
+				return nil, errors.New("Must specify stdin mimetype.")
+			}
+
+			args.pipedMimetype = osArgs[i+1]
+			i++
+
+			continue
+		}
+
+		if arg == "-h" || arg == "--help" {
+			log.Println(HELP)
+			return nil, nil
+		}
+
+		if arg == "-p" || arg == "--pretty" {
+			args.pretty = true
+			continue
+		}
+
+		if arg == "-v" || arg == "--version" {
+			log.Println("dsq " + Version)
+			return nil, nil
+		}
+
+		if arg == "-c" || arg == "--schema" {
+			args.schema = true
+			continue
+		}
+
+		if arg == "-f" || arg == "--file" {
+			if isLast {
+				return nil, errors.New("Must specify a SQL file.")
+			}
+
+			args.sqlFile = osArgs[i+1]
+			i++
+
+			continue
+		}
+
+		if arg == "--cache" || arg == "-C" {
+			args.cacheSettings.Enabled = true
+			continue
+		}
+
+		if arg == "--cache-file" || arg == "-D" {
+			args.dumpCacheFile = true
+			args.cacheSettings.Enabled = true
+			continue
+		}
+
+		args.nonFlagArgs = append(args.nonFlagArgs, arg)
+	}
+
+	return args, nil
+}
+
 var Version = "latest"
 
 var HELP = `dsq (Version ` + Version + `) - commandline SQL engine for data files
@@ -193,56 +331,57 @@ See the repo for more details: https://github.com/multiprocessio/dsq.`
 func _main() error {
 	log.SetFlags(0)
 	runner.Verbose = false
-	var nonFlagArgs []string
-	stdin := false
-	pretty := false
-	schema := false
-	for _, arg := range os.Args[1:] {
-		if arg == "--verbose" {
-			runner.Verbose = true
-			continue
-		}
 
-		if arg == "-s" || arg == "--stdin" {
-			stdin = true
-			continue
-		}
+	args, err := getArgs()
+	if err != nil {
+		return err
+	}
 
-		if arg == "-h" || arg == "--help" {
-			log.Println(HELP)
-			return nil
-		}
-
-		if arg == "-p" || arg == "--pretty" {
-			pretty = true
-			continue
-		}
-
-		if arg == "-v" || arg == "--version" {
-			log.Println("dsq " + Version)
-			return nil
-		}
-
-		if arg == "-c" || arg == "--schema" {
-			schema = true
-			continue
-		}
-
-		nonFlagArgs = append(nonFlagArgs, arg)
+	// Some commands exit naturally here. Like -h and -v
+	if args == nil {
+		return nil
 	}
 
 	lastNonFlagArg := ""
-	files := nonFlagArgs
+	files := args.nonFlagArgs
 
-	// Empty marker meaning to process from stdin
-	if stdin {
-		files = append([]string{""}, files...)
+	// Grab from stdin into local file
+	mimetypeOverride := map[string]string{}
+	if args.pipedMimetype != "" {
+		pipedTmp, err := ioutil.TempFile("", "dsq-stdin")
+		if err != nil {
+			return err
+		}
+		defer os.Remove(pipedTmp.Name())
+
+		mimetypeOverride[pipedTmp.Name()] = args.pipedMimetype
+
+		_, err = io.Copy(pipedTmp, os.Stdin)
+		if err != nil {
+			return err
+		}
+		pipedTmp.Close()
+		files = append([]string{pipedTmp.Name()}, files...)
 	}
 
-	if len(nonFlagArgs) > 1 {
-		lastNonFlagArg = nonFlagArgs[len(nonFlagArgs)-1]
-		if strings.Contains(lastNonFlagArg, " ") {
-			files = files[:len(files)-1]
+	// If -f|--file not present, query is the last argument
+	if args.sqlFile == "" {
+		if len(files) > 1 {
+			lastNonFlagArg = files[len(files)-1]
+			if strings.Contains(lastNonFlagArg, " ") {
+				files = files[:len(files)-1]
+			}
+		}
+	} else {
+		// Otherwise read -f|--file as query
+		content, err := os.ReadFile(args.sqlFile)
+		if err != nil {
+			return errors.New("Error opening sql file: " + err.Error())
+		}
+
+		lastNonFlagArg = string(content)
+		if lastNonFlagArg == "" {
+			return errors.New("SQL file is empty.")
 		}
 	}
 
@@ -250,13 +389,26 @@ func _main() error {
 		return errors.New("No input files.")
 	}
 
-	projectTmp, err := ioutil.TempFile("", "dsq-project")
-	if err != nil {
-		return err
+	var projectIdHashOrTmp string
+	if args.cacheSettings.Enabled {
+		var err error
+		projectIdHashOrTmp, err = getFilesContentHash(files)
+		if err != nil {
+			log.Printf("Error creating hash for cache mode: %v, defaulting to normal mode", err)
+			args.cacheSettings.Enabled = false
+		}
+	} else {
+		projectTmp, err := ioutil.TempFile("", "dsq-project")
+		if err != nil {
+			return err
+		}
+		defer os.Remove(projectTmp.Name())
+
+		projectIdHashOrTmp = projectTmp.Name()
 	}
-	defer os.Remove(projectTmp.Name())
+
 	project := &runner.ProjectState{
-		Id: projectTmp.Name(),
+		Id: projectIdHashOrTmp,
 		Pages: []runner.ProjectPage{
 			{
 				Panels: nil,
@@ -270,91 +422,48 @@ func _main() error {
 	}
 	defer os.RemoveAll(tmpDir)
 
+	// Does no harm in calculating this even if caching is not on. A few places use this path.
+	cachedPath := filepath.Join(os.TempDir(), "dsq-cache-"+projectIdHashOrTmp+".db")
+	if args.cacheSettings.Enabled {
+		info, err := os.Stat(cachedPath)
+		args.cacheSettings.CachePresent = err == nil && info.Size() != 0
+		if !args.cacheSettings.CachePresent {
+			log.Println("Cache invalid, re-import required.")
+			os.Remove(cachedPath)
+		}
+	}
+
 	ec := runner.NewEvalContext(*runner.DefaultSettings, tmpDir)
-	for i := 0; i < len(files); i++ {
-		file := files[i]
-		panelId := uuid.New().String()
-		resultFile := ec.GetPanelResultsFile(project.Id, panelId)
-		out, err := openTruncate(resultFile)
-		if err != nil {
-			return err
-		}
-
-		readFromStdin := false
-		if file == "" {
-			b, err := ioutil.ReadAll(os.Stdin)
-			if err == nil {
-				if i == len(files)-1 {
-					return errors.New("Expected file extension or mimetype: e.g. cat x.csv | dsq -s csv, or cat x.csv | dsq -s text/csv")
-				}
-				mimetype := files[i+1]
-				if !strings.Contains(mimetype, string(filepath.Separator)) {
-					mimetype = string(runner.GetMimeType("x."+mimetype, runner.ContentTypeInfo{}))
-				}
-
-				if mimetype == "" {
-					return fmt.Errorf("Unknown mimetype or file extension: %s.", mimetype)
-				}
-				i += 1
-
-				cti := runner.ContentTypeInfo{Type: string(mimetype)}
-				err := runner.TransformReader(bytes.NewReader(b), "", cti, out)
-				if err != nil {
-					return err
-				}
-
-				s, err := getShape(resultFile, "")
-				if err != nil {
-					return err
-				}
-
-				project.Pages[0].Panels = append(project.Pages[0].Panels, runner.PanelInfo{
-					ResultMeta: runner.PanelResult{
-						Shape: *s,
-					},
-					Id:   panelId,
-					Name: file,
-				})
-
-				readFromStdin = true
-			}
-
-			continue
-		}
-
-		if !readFromStdin {
-			err := evalFileInto(file, out)
+	// When dumping schema, need to injest even if cache is on.
+	if !args.cacheSettings.CachePresent || !args.cacheSettings.Enabled || lastNonFlagArg == "" {
+		for _, file := range files {
+			panel, err := importFile(project.Id, file, mimetypeOverride[file], ec)
 			if err != nil {
 				return err
 			}
+			project.Pages[0].Panels = append(project.Pages[0].Panels, *panel)
 		}
+	}
 
-		s, err := getShape(resultFile, file)
-		if err != nil {
-			return err
-		}
-
-		project.Pages[0].Panels = append(project.Pages[0].Panels, runner.PanelInfo{
-			ResultMeta: runner.PanelResult{
-				Shape: *s,
-			},
-			Id:   panelId,
-			Name: uuid.New().String(),
-		})
-
-		out.Close()
+	if args.dumpCacheFile {
+		fmt.Println(cachedPath)
+		return nil
 	}
 
 	// No query, just dump transformed file directly out
 	if lastNonFlagArg == "" {
 		resultFile := ec.GetPanelResultsFile(project.Id, project.Pages[0].Panels[0].Id)
-		return dumpJSONFile(resultFile, pretty, schema)
+		return dumpJSONFile(resultFile, args.pretty, args.schema)
 	}
 
 	connector, err := runner.MakeTmpSQLiteConnector()
 	if err != nil {
 		return err
 	}
+
+	if args.cacheSettings.Enabled {
+		connector.DatabaseConnectorInfo.Database.Database = cachedPath
+	}
 	project.Connectors = append(project.Connectors, *connector)
 
 	query := rewriteQuery(lastNonFlagArg)
@@ -370,7 +479,7 @@ func _main() error {
 		},
 	}
 
-	err = ec.EvalDatabasePanel(project, 0, panel, nil)
+	err = ec.EvalDatabasePanel(project, 0, panel, nil, args.cacheSettings)
 	if err != nil {
 		if e, ok := err.(*runner.DSError); ok && e.Name == "NotAnArrayOfObjectsError" {
 			rest := "."
@@ -378,7 +487,7 @@ func _main() error {
 			if err == nil {
 				rest = ": " + files[nth] + "."
 			}
-			return fmt.Errorf("Input is not an array of objects%s", rest)
+			return fmt.Errorf("Input is not an array of objects%s\n", rest)
 		}
 
 		if e, ok := err.(*runner.DSError); ok && e.Name == "UserError" {
@@ -391,7 +500,7 @@ func _main() error {
 	}
 
 	resultFile := ec.GetPanelResultsFile(project.Id, panel.Id)
-	return dumpJSONFile(resultFile, pretty, schema)
+	return dumpJSONFile(resultFile, args.pretty, args.schema)
 }
 
 func main() {
